import numpy as np
import cantera as ct

from constants import * 
from space import space
from thermo import thermo



class electrochem:
    def __init__(self) -> None:

        # Calculate Reversible Cell Potential under standard conditions
        h2 = ct.Hydrogen()
        o2 = ct.Oxygen()
        h2o = ct.Water()
        h2.TP = STD_TEMPERATURE, STD_PRESSURE
        o2.TP = STD_TEMPERATURE, STD_PRESSURE
        h2o.TP = STD_TEMPERATURE, STD_PRESSURE

        dG = h2.gibbs_mole + 0.5 * o2.gibbs_mole - h2o.gibbs_mole # J/kmol
        self.U_rev_0 = - dG / (2 * FARRADAY * 1e3)


    def calcCurrentDensity(self, space:space, u_cell:float, i_last:np.ndarray, thermo:thermo) -> np.ndarray:
        # Calculates the current density in the cells
        # Args:
        #   space: the space object
        #   u_cell: applied cellvoltage
        #   i_last: cell current from the last timestep in A/m^2
        # Returns: 
        #   i:  current density


        u_rev = self.calcReversibleCellPotential(space)
        u_overpotential = self.calcOverpotentials(space, i_last)

        u_cell_last = u_rev - u_overpotential[0] - u_overpotential[1] - u_overpotential[2] - u_overpotential[3] - u_overpotential[4]
        space.voltage = u_cell_last

        i = u_cell / u_cell_last * i_last # A/m^2
        i[i<0.001] = 0.001 # make shure that not trapped at 0
        #i[:,-1] = 0 # backside of last EC col -> no EC in next col

        self.heat_overpotential = np.zeros_like(space.tempspace.t_c)
        self.heat_overpotential[np.array(INDEX_ACL)-1, :] = np.multiply((u_overpotential[0] + u_overpotential[3]), i * CELLSIZE_Z * space.dy[INDEX_ACL,1:COLPTS+1], out=np.zeros_like(u_overpotential[0]),where=(i!=0)) 
        self.heat_overpotential[np.array(INDEX_CCL)-1, :] = np.multiply((u_overpotential[1] + u_overpotential[4]), i * CELLSIZE_Z * space.dy[INDEX_CCL,1:COLPTS+1], out=np.zeros_like(u_overpotential[0]),where=(i!=0))
        self.heat_overpotential[np.array(INDEX_MEMBRANE)-1, :] = np.multiply(u_overpotential[2], i * CELLSIZE_Z * space.dy[INDEX_MEMBRANE,1:COLPTS+1], out=np.zeros_like(u_overpotential[0]),where=(i!=0)) 
        # add entropic reaction heat 
        mean_t = np.mean(np.array([space.tempspace.t[INDEX_ACL,1:COLPTS+1], space.tempspace.t[INDEX_CCL,1:COLPTS+1]]))
        s_h2o, s_h2, s_o2 = thermo.calcEntropy(mean_t)
        entropic_heat = mean_t * (s_h2o - s_h2 - .5*s_o2) / 1000 *  (i * CELLSIZE_Z * space.dy[INDEX_ACL,1:COLPTS+1]) / (2*FARRADAY) # W/cell
        self.heat_overpotential[np.array(INDEX_ACL)-1, :] += (entropic_heat/2)  # Divide half half between anode and cathode as proposed by Bock.2020
        self.heat_overpotential[np.array(INDEX_CCL)-1, :] += (entropic_heat/2) 

        # apply half of source to cell before and half of source to cell and half to cell after to have source between the cells (in the layer of the EC)            
        self.heat_overpotential *= .5
        self.heat_overpotential[1:,:] += self.heat_overpotential[:-1,:]

        
        if DEBUG:
            try:
                print(f"u_rev: {u_rev.min()} - {u_rev.max()}V\nu_cell_last: {np.nanmin(u_cell_last)} - {np.nanmax(u_cell_last)}V\nu_cell_set: {u_cell}V\ni: {i[i!=0].min()/1e4} - {i.max()/1e4}A/cm2\nmu_act_a: {u_overpotential[0][np.isfinite(u_overpotential[0])].min()} - {u_overpotential[0].max()}V\nmu_act_c:  {u_overpotential[1][np.isfinite(u_overpotential[1])].min()} - {u_overpotential[1][np.isfinite(u_overpotential[1])].max()}V\nmu_ohm:  {u_overpotential[2][u_overpotential[2]!=0].min()} - {u_overpotential[2].max()}V\nmu_mt:  {u_overpotential[3][u_overpotential[3]!=0].min()} - {u_overpotential[3].max()}V")
            except:
                pass
        return i
    
    def calcHenrysCoefficient(self, T:np.ndarray, fluid:str) -> np.ndarray:
        # Calculates the temperature depentent value of Henry's coefficient of the given gass in water
        # Args:
        #   T: the temperatures
        #   fluid: the type of fluid ('H2' for Hydrogen, 'O2' for Oxygen )
        # Returns: 
        #   henry:  henry coefficients in mol /m^3 / Pa

        henry = np.zeros_like(T)

        if fluid == "H2":
            # Using equation from Battino and Wilhelm (given in Ito.2011) and devide by 18 mol(H2O)/cm^3 and 1e5 Pa per bar -> bullshit values
            henry = np.exp(-48.1611 + (5528.45 / T) + 16.8893 * np.log(T / 100)) / 18 / 1e5 *1e6
            # henry[T < 45 + 273.15] = (7.9*1e6 * np.exp(-545 / T[T< 45 + 273.15]) * (1 + 0.000071*(PRESSURE/101325)**3))**-1 / 101325 * 1e6
            # henry[T > 45 + 273.15] = (8.34*1e5 * np.exp(170 / T[T> 45 + 273.15]) * (1 + 0.000071*(PRESSURE/101325)**3))**-1 / 101325 * 1e6

            # Using eqation from Sander.2015:
           # henry = 7.7e-6 * np.exp(500 * ((1/T) - (1/298.15)))

        elif fluid == "O2":
            # Using equation from Battino (given in Ito.2011) and devide by 18 mol(H2O)/cm^3 and 1e5 Pa per bar -> bullshit values
            henry = np.exp(-66.73538 + (8747.547 / T) + 24.45264 * np.log(T / 100)) / 18 / 1e5 * 1e6
            # henry[T < 45 + 273.15] = (1.34*1e8 * np.exp(-1540 / T[T< 45 + 273.15]))**-1 / 101325 * 1e6
            # henry[T > 45 + 273.15] = (5.08*1e6 * np.exp(-500 / T[T> 45 + 273.15]) )**-1 / 101325 * 1e6

            # Using eqation from Sander.2015:
            #henry = 1.3e-5 * np.exp(1500 * ((1/T) - (1/298.15)))

        else:
            raise Exception("Unknown Fluid!")
        
        return henry


    def calcOverpotentials(self, space:space, i:np.ndarray,bock:bool=False) -> [np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        # Calculates the overpotential of the cells
        # Args: 
        #   space: the space object
        #   i: cell current density in section in A/m2
        # Returns: 
        #   mu_act_anode: anodic activation overpotentials
        #   mu_act_cathode: cathodic activation overpotentials
        #   mu_ohm: ohmic overpotentials
        #   mu_mt: mass transfer overpotentials 

        T_cells_m = space.tempspace.t[INDEX_MEMBRANE,1:COLPTS+1] # temperature membranes in K 
        T_cells_a = space.tempspace.t[INDEX_ACL,1:COLPTS+1] # temperature anodes in K 
        T_cells_c = space.tempspace.t[INDEX_CCL,1:COLPTS+1] # temperature cathodes in K 

        ### Calculate activation overpotential ##
        ## Accoding to Krenz ##
        #Anodic using tafel kinetics
        alpha_ct = ALPHA_CT_REF_A * np.exp(E_CT_A*(1/333.15 - 1/T_cells_a))  # temperature dependent charge transfer coefficient (using Arrhenius approach to include temperature effects)
        tafel_slope = (R*T_cells_a) / (2 * alpha_ct * FARRADAY)
        i_0_a = I_0_A_REF * np.exp(E_ACD/R * (1/353.15 - 1/T_cells_a))# Anodic exchange current density (using Arrhenius approach)

        mu_act_anode = tafel_slope * np.log((i) / i_0_a)
        mu_act_anode[mu_act_anode<0] = 0

        # Cathodic (accoring to Abdin.2015) -> much smaller than anode -> influence of temperature neglected
        #alpha_ct = ALPHA_CT_REF_C * np.exp(E_CT_C*(1/333.15 - 1/T_cells_c))  # temperature dependent charge transfer coefficient (using Arrhenius approach to include temperature effects)
        tafel_slope = (R*T_cells_c) / (ALPHA_CT_REF_C * FARRADAY)
        i_0_c = I_0_C_REF * np.exp(E_CCD/R * (1/298.15 - 1/T_cells_c))# Anodic exchange current density (using Arrhenius approach)

        mu_act_cathode = tafel_slope * np.arcsinh((i) / i_0_c)
        mu_act_cathode[mu_act_cathode<0] = 0


        ### Calculate Ohmic losses ##
        # membrane electric resistance 
        el_cond_membrane = (K_MEM * LAMBDA - 0.326) * np.exp(E_MEM * (1/303 - 1/T_cells_m))# elecric conductivity of the membrane, calculated using approch of Springer et al. 
        r_membrane =  THICKNESS_MEMBRANE / el_cond_membrane # WAS THE WRONG WAY ARROUND IN KRENZ.2023

        # contact resistance assumed to be constant (R_CONTACT)

        # sum up 
        mu_ohm = i * (r_membrane + R_CONTACT) 


        ### Mass transfer losses ## 
        # Due to supersaturation concentrations of products higher in CLs because of mass transfere resistance out of CLs  
        # Use mass flow in flowfild to approximate composition in CL 

        #IMPLEMENTATION FOLLOWING KRENZ.2023 -> SEEMS BIT STRANGE TO ME (MUCH TO LOW ETC.) -> USE IMPLEMENTATION FOLLOWING Toghyani.2019
        # Calculate saturated concentrations of H2 and O2     
        # m_y = space.flowspace.m_y.astype(float, copy=False)
        # m_y_h2o_g = space.flowspace.m_y_h2o_g.astype(float, copy=False)
        c_c_h2_sat = PRESSURE * self.calcHenrysCoefficient(T_cells_c,'H2') # concentration of H2 at the cathode under actual conditions in mol/m3
        c_a_o2_sat = PRESSURE * self.calcHenrysCoefficient(T_cells_a,'O2') # concentration of O2 at the anode under actual conditions in mol/m3
        
        # Diffusivity of Oxygen inside the PEM in m2/s
        # D_eff_O2_H2O_g = POROSITY_PTL_A**(1.5) * D_O2_H2O_G # Diffusivity in PTL from bulk diffusivity of oxygen in water vapor by Bruggeman's correction (Wu.2007) 
        # D_eff_O2_H2O_l = POROSITY_PTL_A**(1.5) * D_O2_H2O_L # Diffusivity in PTL from bulk diffusivity of oxygen in liquid water by Bruggeman's correction (Wu.2007)
        # D_eff_H2_H2O_g = POROSITY_PTL_C**(1.5) * D_H2_H2O_G # Diffusivity in PTL from bulk diffusivity of hydrogen in water vapor by Bruggeman's correction (Wu.2007)
        # D_eff_H2_H2O_l = POROSITY_PTL_C**(1.5) * D_H2_H2O_L # Diffusivity in PTL from bulk diffusivity of hydrogen in water vapor by Bruggeman's correction (Wu.2007)
        # H_o2 = self.calcHenrysCoefficient(T_cells_a,'O2')# Henry constant for dissolution of oxygen in liquid water 
        # #H_o2 = np.exp(-(666/T_cells_a)+14.1) # in atm cm3 /mol (from Baschuk.2000, who proposed eqauation for D_eff_O2)
        # v_o2 = np.abs(m_y[INDEX_ANODE_FLOWFIELD,1:COLPTS+1,2]) * R * T_cells_a / (PRESSURE / MOLAR_MASS_O2)   # volume flow oxygen in anode CL using iG
        # v_h2o_g = np.abs(m_y_h2o_g[INDEX_ANODE_FLOWFIELD,1:COLPTS+1]) * R * T_cells_a / (PRESSURE / MOLAR_MASS_H2O) # volume flow gasous h2o in anode CL using iG
        # v_h2o_l = (np.maximum(np.abs(m_y[INDEX_ANODE_FLOWFIELD,1:COLPTS+1,0])-np.abs(m_y_h2o_g[INDEX_ANODE_FLOWFIELD,1:COLPTS+1]),0)) / 1000 # volume flow liquid h2o in anode CL using density liquid water (1000 kg / m3)
        # f_w_a = v_h2o_l / (v_h2o_l + v_h2o_g + v_o2)  # Volume fraction of void region in anode that is flodded by liquid water (0-10% in Wu.2017); Assuming euqal flow speed of liquid and gas 
        # if DEBUG: 
        #     #print(f"Liquid water volume fraction in ACL: {f_w_a}")
        #     pass

        # D_eff_O2 = ((1-f_w_a)/D_eff_O2_H2O_g*(R*T_cells_a/H_o2) + f_w_a/D_eff_O2_H2O_l)**-1  # effective diffusivity for o2 in mixture present at anode side (Wu.2007, Baschuk.2000) 
        # Assumed that only water vapor at cathode -> here no combination of liquid and gas diffusivity needed 

        # Supersaturation concentrations
        # c_a_O2_supersat = (i/(4*FARRADAY) + K_L * c_a_o2_sat) / (K_L + (D_eff_O2_H2O_l/THICKNESS_MEMBRANE)) # last term diffusion through membrane (Trinke.2017)
        # c_c_H2_supersat = (i/(2*FARRADAY) + K_L * c_c_h2_sat) / (K_L + (D_eff_H2_H2O_l/THICKNESS_MEMBRANE))
        c_a_O2_supersat = (i/(4*FARRADAY) + K_L * c_a_o2_sat) / (K_L) # Neglecting diffusion through membrane, because wrongly used in Krenz.2023 as diffusion through CLs, but thats already included in K_L (Trinke.2017)
        c_c_H2_supersat = (i/(2*FARRADAY) + K_L * c_c_h2_sat) / (K_L)
        c_a_O2_supersat[c_a_O2_supersat<c_a_o2_sat] = c_a_o2_sat[c_a_O2_supersat<c_a_o2_sat]
        c_c_H2_supersat[c_c_H2_supersat<c_c_h2_sat] = c_c_h2_sat[c_c_H2_supersat<c_c_h2_sat]

        # Calc  mass transfer overpotential 
        mu_mt_a = (R * T_cells_a) / (4 * FARRADAY) * np.log(np.divide(c_a_O2_supersat,c_a_o2_sat,out=np.zeros_like(c_a_O2_supersat),where=(c_a_o2_sat!=0)))
        mu_mt_c = (R * T_cells_c) / (2 * FARRADAY) * np.log(np.divide(c_c_H2_supersat,c_c_h2_sat,out=np.zeros_like(c_c_H2_supersat),where=(c_c_h2_sat!=0)))
        
        # Overpotentials from Bock.2020
        if bock:
            mu_ohm = i/8.7*THICKNESS_MEMBRANE
            mu_act_anode = 0.64 + 0.07 * np.log(i/10000)
            mu_act_cathode = 0.01 + 0.03 * np.log(i/10000)
            mu_mt_a=0
            mu_mt_c=0




        return [mu_act_anode, mu_act_cathode, mu_ohm, mu_mt_a, mu_mt_c]





    def calcReversibleCellPotential(self,space:space) -> np.ndarray:
        # Calculates the reversible cell potential using the Nernst equation
        # Args: 
        #   space: the space object
        # Returns: 
        #   U_rev: ndarray with the reversible cell potentials of all y-sections of the cells in the stack

        # m_y = space.flowspace.m_y.astype(float, copy=False)
        # m_y_total_anode = np.sum(np.abs(m_y[INDEX_ANODE_FLOWFIELD,1:COLPTS+1,:]),axis=2)
        # m_y_total_cathode = np.sum(np.abs(m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,1:COLPTS+1,:]),axis=2)
        # p_o2 = np.divide(np.abs(m_y[INDEX_ANODE_FLOWFIELD,1:COLPTS+1,2]) * PRESSURE, m_y_total_anode, out=np.zeros_like(m_y[INDEX_ANODE_FLOWFIELD,1:COLPTS+1,2]), where=m_y_total_anode!=0)
        # p_h2 = np.divide(np.abs(m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,1:COLPTS+1,1]) * PRESSURE,  m_y_total_cathode, out=np.zeros_like(m_y[INDEX_CATHODE_FLOWFIELD,1:COLPTS+1,1]), where=m_y_total_cathode!=0)


        T_cells_a = space.tempspace.t[INDEX_ACL,1:COLPTS+1]
        T_cells_c = space.tempspace.t[INDEX_CCL, 1:COLPTS+1]
        T_cells_membrane = space.tempspace.t[INDEX_MEMBRANE,1:COLPTS+1]
        T_cells = np.mean([T_cells_a,T_cells_c,T_cells_membrane], axis=0)

        # calculate concentrations of gasses
        c_c_h2_std  = STD_PRESSURE * self.calcHenrysCoefficient(np.array(STD_TEMPERATURE),'H2') # concentration of H2 at the cathode under standard conditions in mol/m3
        c_a_o2_std  = STD_PRESSURE * self.calcHenrysCoefficient(np.array(STD_TEMPERATURE),'O2') # concentration of O2 at the anode under standard conditions in mol/m3
        c_c_h2 = PRESSURE * self.calcHenrysCoefficient(T_cells_c,'H2') # concentration of H2 at the cathode under actual conditions in mol/m3
        c_a_o2 = PRESSURE * self.calcHenrysCoefficient(T_cells_a,'O2') # concentration of =2 at the anode under actual conditions in mol/m3


        # Nernst equation (actually activityCoefficient * concentaration but because acitityCoefficient~1 for low concentrations neglected )
        U_rev = self.U_rev_0 - R * T_cells/ (2 * FARRADAY) * np.log((c_a_o2/c_a_o2_std)**0.5 * (c_c_h2/c_c_h2_std))
        #U_rev = self.U_rev_0 - R * T_cells/ (2 * FARRADAY) * np.log((p_h2/STD_PRESSURE)**0.5 * (p_o2/STD_PRESSURE) / 1)

        
        
        return U_rev
    
    def calcThermoneutralCellPotential(self,space:space) -> np.ndarray:
        # Calculates the thermoneutral cell potential
        # Args: 
        #   space: the space object
        # Returns: 
        #   U_thn: ndarray with the thermoneutral cell potentials of all y-sections of the cells in the stack

        T_cells_a = space.tempspace.t[INDEX_ACL,1:COLPTS+1]
        T_cells_c = space.tempspace.t[INDEX_CCL, 1:COLPTS+1]
        T_cells_membrane = space.tempspace.t[INDEX_MEMBRANE,1:COLPTS+1]
        T_cells = np.mean([T_cells_a,T_cells_c,T_cells_membrane], axis=0)

        U_thn = np.zeros_like(T_cells)
        
        h2 = ct.Hydrogen()
        o2 = ct.Oxygen()
        h2o = ct.Water()

        for x in range(np.size(T_cells,0)):
            for y in range(np.size(T_cells,1)):
                h2.TP = T_cells[x,y], STD_PRESSURE
                o2.TP = T_cells[x,y], STD_PRESSURE
                h2o.TP = T_cells[x,y], STD_PRESSURE

                dH = h2.enthalpy_mole + 0.5 * o2.enthalpy_mole - h2o.enthalpy_mole # J/kmol
                U_thn[x,y] = - dH / (2 * FARRADAY * 1e3)

        return U_thn