import json
import msgpack
import msgpack_numpy
msgpack_numpy.patch() # Adds ability to serialize np arrays to msgpack
import numpy as np
import os 

from constants import *
import constants
from space import space

def makeDirectory(directory:str=RESULTS_DIRECTORY, storeParameters:bool=True, wipe:bool=WIPE_DIRECTORY):
        # Creates a directory to write results to 
        # Args:
        #   directory: directory to store results in
        #   storeParameters: if parameters should be stored as json in the created directory
        #   wipe: if existing files in directory should be deleted before saving new files

        # If directory does not exist, create it 
        if not os.path.isdir(directory):
                os.makedirs(directory, exist_ok=False)
        else:
            if wipe:
                #remove all files in directory
                conf = input(f'Do you really want to remove all files from "{directory}"? (Y/N)')
                if conf == 'Y' or conf == 'y':
                        os.chdir(directory)
                        filelist = os.listdir()
                        for file in filelist:
                                if os.path.isfile(file):
                                        # is a file and not a folder
                                        os.remove(file)
        
        # Store all parameters as json 
        if storeParameters:
                filename = f"parameters.json"
                path = os.path.join(directory,filename)
                consts = dict()
                for name, val in globals().items():
                        if name.isupper(): # excludes internal variables of module
                                if isinstance(val,np.ndarray):
                                        consts[name] = val.tolist()
                                elif isinstance(val, range):
                                       consts[name] = list(val)
                                else:
                                        consts[name] = val
                        
                with open(path,'w') as f:
                        json.dump(consts,f)

        
        os.chdir(directory)

def readParameters(directory:str) -> dict:
        # Reads the parameters from the given file 
        # Args:
        #   directory:  path to directory where the parameters.json is stored
        # Returns:
        #   params: dict with all parameters 
        path = os.path.join(directory,'parameters.json')
        with open(path, 'r') as f:
                params = json.load(f)
        return params


def readResultDictory(directory:str) -> np.ndarray:
        # Checks available iterations in directory
        # Args:
        #   directory:  path to directory to read from
        # Returns:
        #   sorted_iterations: the iterations available in the given directory as np array of ints


        os.chdir(directory)#Go through files in the directory and store filenames
        filenames=[]
        iterations=[]
        for datafile in os.listdir(directory) :
                if os.path.isfile(os.path.join(directory,datafile)):
                        if "PUV" in datafile and datafile[-8:] == ".msgpack":
                                filenames.append(datafile)
                                no_ext_file=datafile.replace(".msgpack","").strip()
                                iter_no=int(no_ext_file.split("V_")[-1])
                                iterations.append(iter_no)#Discern the final iteration and interval
        initial_iter=np.amin(iterations).astype(int)          
        final_iter=np.amax(iterations).astype(int)
        inter=(final_iter - initial_iter)/(len(iterations)-1)
        number_of_frames=len(iterations)
        sorted_iterations=np.sort(iterations)
        return sorted_iterations

def readFromFile(iteration:int, directory:str, variable:str) -> [np.ndarray, str]:
        # Reads data of iteration from file
        # Args:
        #   iteration: iteration that should be read
        #   directory:  directory of the file
        # returns:
        #   data: values of the stored variable
        #   name: name of the stored variable 

         #Set filename and path according to given iteration
        filename="PUV_{0}.msgpack".format(iteration)
        filepath=os.path.join(directory,filename)    #Load text file as numpy array

        with open(filepath, 'rb') as f:
              byte_data = f.read()
        
        data = msgpack.unpackb(byte_data)
        
        # return data['p_c'],data['u_c'],data['v_c']
        return data[variable]['data'].astype(float,copy=True), data[variable]['name']
def writeToFile(space:space, iteration, isFinal=False) -> None:
        # Creates a directory to write results to 
        # Args:
        #   space:  the space object
        #   iteration: current iteration
        #   isFinal: use True if final result of simulation
        if(iteration%WRITE_INTERVALL==0 or isFinal):
                filename = f"PUV_{iteration}.msgpack"
                path = os.path.join(RESULTS_DIRECTORY,filename)
                data = {
                        't_c':{
                                'data': space.tempspace.t_c-273.15,
                                'name': 'Temperature [°C]',
                        },
                        'm_x_h2o2':{
                                'data': space.flowspace.m_x_c[:,:,1:2],
                                'name': 'Massflow H2 and O2 in x-direction [Kg/s]',
                        },
                        'm_y_h2o2':{
                                'data': space.flowspace.m_y_c[:,:,1:2],
                                'name': 'Massflow H2 and O2 in y-direction [Kg/s]',
                        },
                        'm_x_h2o':{
                                'data': space.flowspace.m_x_h2o_c[:,:],
                                'name': 'Massflow H2O in x-direction [Kg/s]',
                        },                        
                        'm_y_h2o':{
                                'data': space.flowspace.m_y_h2o_c,
                                'name': 'Massflow H2O in y-direction [Kg/s]',
                        },
                        'i':{
                                'data': space.current_desity[:,:]/1e4,     
                                'name': 'Current density [A/cm^2]',
                        },
                        'u':{
                                'data': space.voltage[:,:],
                                'name': 'Cell voltage [V]',
                        },   
                        'heat_flux_x':{
                                'data': space.heat_flux[:,:,0],
                                'name': "Heat flux in x-direction [W/m^2]",
                        },              
                        'heat_flux_y':{
                                'data': space.heat_flux[:,:,1],
                                'name': "Heat flux in y-direction [W/m^2]",
                        },              
                      }
                with open(path,'wb') as f:
                        packed = msgpack.packb(data)
                        f.write(packed)


