import numpy as np
import cantera as ct 

from constants import *
from space import space

class thermo: 
    # is responsible for thermodynamic calculations

    def __init__(self) -> None:
        self.__h2o__ = ct.Water()
        self.__h2__ = ct.Hydrogen()
        self.__o2__ = ct.Oxygen()
        self.T_ref = 343.15 # K (70°C)

        #get enthalpy at reference temperature
        self.__h2o__.TP = self.T_ref, PRESSURE
        self.__h2__.TP = self.T_ref, PRESSURE
        self.__o2__.TP = self.T_ref, PRESSURE


        self.h_h2o = self.__h2o__.enthalpy_mass # J/Kg
        self.h_h2 = self.__h2__.enthalpy_mass # J/Kg
        self.h_o2 = self.__o2__.enthalpy_mass # J/Kg

        self.s_h2o = self.__h2o__.entropy_mole # J/kmol/K
        self.s_h2 = self.__h2__.entropy_mole # J/kmol/K
        self.s_o2 = self.__o2__.entropy_mole # J/kmol/K


        self.cv_h2o = self.__h2o__.cv_mass # J/Kg/K
        self.cp_h2 = self.__h2__.cp_mass # J/Kg/K
        self.cp_o2 = self.__o2__.cp_mass # J/Kg/K
        self.cv_h2o_mole = self.__h2o__.cv_mole # J/Kg/K
        self.cv_h2_mole = self.__h2__.cv_mole # J/Kg/K
        self.cv_o2_mole = self.__o2__.cv_mole # J/Kg/K

        self.h_sat_steam = []

        for T in range(5,150,10):
            self.__h2o__.TQ = 273.15 + T, 1 
            self.h_sat_steam.append(self.__h2o__.enthalpy_mass)
        self.h_sat_steam = np.array(self.h_sat_steam)
        

    def approxVaporEnthalpy(self, T:np.ndarray) -> np.ndarray:
        # approximates the enthalpy of saturated water vapor at the given temperature 
        # Args:
        #   T: the temperature in K
        # Returns:
        #   h: the enthalpy in J/Kg
        if T.max() > 145 + 273.15 or np.any(np.isnan(T)):
            raise Exception(f"Maximum temperature {(T.max()-273.15)}°C out of valid range (> 145°C)!")
        if T.min() < 5 + 273.15:
            raise Exception(f"Minimal temperature {(T.min()-273.15)}°C out of valid range (< 5°C)!")

        idxT = np.floor((T-278.15)/10).astype(int)
        h = (self.h_sat_steam[idxT+1] - self.h_sat_steam[idxT])/10 * (T-(idxT*10+278.15)) + self.h_sat_steam[idxT]
        return h
    
    def calcEntropy(self, temp:np.ndarray) -> [np.ndarray, np.ndarray, np.ndarray]:
        # Calculates the molar entropy of water, hydrogen and oxygen at the given temperatures
        # Args:
        #   temp: temperatures where entropy should be calclated in K
        # Returns:
        #   s_h2o: entropy water in J/kmol/K
        #   s_h2:   entropy hydrogen in J/kmol/K
        #   s_o2:   entropy oxygen in J/kmol/K

        s_h2o = self.s_h2o + self.cv_h2o_mole * np.log(temp/self.T_ref)       
        s_h2 = self.s_h2 + self.cv_h2_mole * np.log(temp/self.T_ref)       
        s_o2 = self.s_o2 + self.cv_o2_mole * np.log(temp/self.T_ref)       

        return s_h2o, s_h2, s_o2 




    def calcHeatSourcesFluidsFullyEquilibriated(self, space:space) -> np.ndarray:
        # calculate the heat sources from temperature change of the fluids assuming same temperature in solid and fluids 
        # Args:
        #   space: space object
        # Returns:
        #   source: ndarray containing the sources through change in temperature of the fluids in W/cell
        
        m_x = space.flowspace.m_x.astype(float,copy=False)
        m_x_h2o_g = space.flowspace.m_x_h2o_g.astype(float, copy=False)
        
        m_y = space.flowspace.m_y.astype(float,copy=False)
        m_y_h2o_g = space.flowspace.m_y_h2o_g.astype(float, copy=False)
        T = space.tempspace.t.astype(float,copy=False)

        # Enthaply of fluids at meshpoints
        dT = T-self.T_ref
        _m_x_h2o_g = (np.minimum(np.abs(m_x_h2o_g[:,:]), np.abs(m_x[:,:,0])) * np.sign(m_x[:,:,0]))
        _m_x_h2o_l = ((np.abs(m_x[:,:,0]) - np.abs(_m_x_h2o_g[:,:])) * np.sign(m_x[:,:,0]))
        H_h2o_x_l = _m_x_h2o_l * (self.h_h2o+dT*self.cv_h2o) # J/s
        H_h2o_x_g = _m_x_h2o_g * self.approxVaporEnthalpy(T)
        space.flowspace.m_x_h2o_c[:,:,0] =  _m_x_h2o_l[1:ROWPTS+1,1:COLPTS+1]
        space.flowspace.m_x_h2o_c[:,:,1] =  _m_x_h2o_g[1:ROWPTS+1,1:COLPTS+1]
        H_h2_x = m_x[:,:,1] * (self.h_h2+dT*self.cp_h2) # J/s
        H_o2_x = m_x[:,:,2] * (self.h_o2+dT*self.cp_o2) # J/s
        _m_y_h2o_g = (np.minimum(np.abs(m_y_h2o_g[:,:]), np.abs(m_y[:,:,0])) * np.sign(m_y[:,:,0]))
        _m_y_h2o_l = ((np.abs(m_y[:,:,0]) - np.abs(_m_y_h2o_g[:,:])) * np.sign(m_y[:,:,0]))
        H_h2o_y_l = _m_y_h2o_l * (self.h_h2o+dT*self.cv_h2o) # J/s
        H_h2o_y_g = _m_y_h2o_g * self.approxVaporEnthalpy(T)
        space.flowspace.m_y_h2o_c[:,:,0] =  _m_y_h2o_l[1:ROWPTS+1,1:COLPTS+1]
        space.flowspace.m_y_h2o_c[:,:,1] =  _m_y_h2o_g[1:ROWPTS+1,1:COLPTS+1]
        H_h2_y = m_y[:,:,1] * (self.h_h2+dT*self.cp_h2) # J/s
        H_o2_y = m_y[:,:,2] * (self.h_o2+dT*self.cp_o2) # J/s

        # Change in enthalpy through reaction (is substitued from enthalpy change in ACL)
        n_reaction = space.current_desity*CELLAREA_SEGMENT/(2 * FARRADAY)  # molar flow reaction
        dH_reaction = (n_reaction*MOLAR_MASS_H2*(self.h_h2 + dT[INDEX_ACL,1:COLPTS+1]*self.cp_h2) + .5*n_reaction*MOLAR_MASS_O2*(self.h_o2 + dT[INDEX_ACL,1:COLPTS+1]*self.cp_o2) - n_reaction*MOLAR_MASS_H2O*(self.h_h2o + dT[INDEX_ACL,1:COLPTS+1]*self.cv_h2o) )

        # Enthalpy change in cells in x and y direction
        dHx = ((H_h2o_x_l[:ROWPTS+1, 1:COLPTS+1] - H_h2o_x_l[1:,1:COLPTS+1]) + (H_h2o_x_g[:ROWPTS+1, 1:COLPTS+1] - H_h2o_x_g[1:,1:COLPTS+1])) + (H_h2_x[:ROWPTS+1, 1:COLPTS+1] - H_h2_x[1:,1:COLPTS+1]) + (H_o2_x[:ROWPTS+1, 1:COLPTS+1] - H_o2_x[1:,1:COLPTS+1])
        dHy = ((H_h2o_y_l[1:ROWPTS+1, :COLPTS+1] - H_h2o_y_l[1:ROWPTS+1,1:]) + (H_h2o_y_g[1:ROWPTS+1, :COLPTS+1] - H_h2o_y_g[1:ROWPTS+1,1:]))+ (H_h2_y[1:ROWPTS+1, :COLPTS+1] - H_h2_y[1:ROWPTS+1,1:]) + (H_o2_y[1:ROWPTS+1, :COLPTS+1] - H_o2_y[1:ROWPTS+1,1:])
        # Substract reaction enthalpy 
        dHx[INDEX_ACL,:] += dH_reaction
        # Remove change in enthalpy that is only caused by change in flow-direction
        dHy[np.array(INDEX_ANODE_FLOWFIELD)-1,1:] += dHx[INDEX_ANODE_FLOWFIELD,:]
        dHy[np.array(INDEX_CATHODE_FLOWFIELD_BACKSIDE)-1,1:] += dHx[INDEX_CATHODE_FLOWFIELD,:]
        dHx[INDEX_ANODE_FLOWFIELD,:] -= dHx[INDEX_ANODE_FLOWFIELD,:]
        dHx[INDEX_CATHODE_FLOWFIELD,:] -= dHx[INDEX_CATHODE_FLOWFIELD,:]
        
        # distribute heatflux between neighboring cells according to cellsize
        # dH = ((space.area[1:ROWPTS+1,1:COLPTS+1]/(space.area[1:ROWPTS+1,1:COLPTS+1]+space.area[2:,1:COLPTS+1])) * dHx[1:ROWPTS+1,:] + (space.area[1:ROWPTS+1,1:COLPTS+1]/(space.area[1:ROWPTS+1,1:COLPTS+1]+space.area[0:ROWPTS,1:COLPTS+1])) * dHx[0:ROWPTS,:]) 
        dH = .5 * dHx[1:ROWPTS+1,:] + .5 * dHx[0:ROWPTS,:]
        dH[np.array(INDEX_ANODE_FLOWFIELD)-1,:] += dHy[np.array(INDEX_ANODE_FLOWFIELD)-1,:COLPTS]
        if PARALLEL_FLOW:
            dH[np.array(INDEX_CATHODE_FLOWFIELD_BACKSIDE)-1,:] += dHy[np.array(INDEX_CATHODE_FLOWFIELD_BACKSIDE)-1,:COLPTS]        
        else:
            dH[np.array(INDEX_CATHODE_FLOWFIELD_BACKSIDE)-1,:] += dHy[np.array(INDEX_CATHODE_FLOWFIELD_BACKSIDE)-1,1:]
       
        return dH
        # dH_A = dH / space.area[1:ROWPTS+1,1:COLPTS+1] # W/m^2
        # return dH_A
    
    def calcHeatFlowToAmbient(self, space:space) -> np.ndarray:
        # calculate the heat flux from the end-plates to ambient (0 if no Endplates)  
        # Args:
        #   space: space object
        # Returns:
        #   source: ndarray containing the source term through heat flow from end-plates to ambient in W/cell related to cellarea in simulation direction

        source = np.zeros((ROWPTS,COLPTS))
        if HAS_ENDPLATE:
            source[0,:] = - ALPHA_ENDPLATE_AMBIENT * CELLSIZE_Z * space.dy[1,1:COLPTS+1] * (space.tempspace.t[1,1:COLPTS+1] - T_AMBIENT) # left end-plate (cellsizes to relate heat flux to cellarea in simlation dimension)
            source[-2,:] = - ALPHA_ENDPLATE_AMBIENT * CELLSIZE_Z * space.dy[1,1:COLPTS+1] * (space.tempspace.t[-2,1:COLPTS+1] - T_AMBIENT) # right end-plate (use temp at right side of end plate but apply source to left side with thickness of end-plate)
        
        return source


    def calcMassFlowGasousWaterInH2O2(self, space:space, m_dry_h2:np.ndarray, m_dry_o2:np.ndarray) -> np.ndarray:
        # calculate the saturated mass flow of gasous water in the gasses
        # Args:
        #   space: space object
        #   m_dry_h2: the total mass flow of H2 without water
        #   m_dry_o2: the total mass flow of O2 without water
        # Returns:
        #   m_h2o_in_gas: ndarray containing the mass flow of water in H2 and O2 in Kg/s
        h2_in_mixture = np.divide(m_dry_h2, (m_dry_h2 + m_dry_o2), out=np.ones_like(m_dry_h2), where=(m_dry_h2 + m_dry_o2)!=0)

        p_sat = self.calcSaturationPressureWater(space.tempspace.t)
        m_h2o_in_h2 = np.divide((MOLAR_MASS_H2O / (MOLAR_MASS_H2)) * p_sat * m_dry_h2, ((PRESSURE - p_sat)*h2_in_mixture), out=np.zeros_like(p_sat), where=(h2_in_mixture!=0))
        m_h2o_in_o2 = np.divide((MOLAR_MASS_H2O / (MOLAR_MASS_O2)) * p_sat * m_dry_o2, ((PRESSURE - p_sat)*(1-h2_in_mixture)), out=np.zeros_like(p_sat), where=(h2_in_mixture!=1))
        m_h2o_in_h2[m_dry_h2==0] = m_h2o_in_o2[m_dry_h2==0] # take values from o2 where no h2 in mixture
        m_h2o_in_h2[p_sat > PRESSURE] = 1e10 # all steam
        return m_h2o_in_h2

    def calcSaturationPressureWater(self, T:np.ndarray) -> np.ndarray:
        # calculate the saturation pressure of water using the antoine equation and parameters from Roizard.2019 (added conversion to from °C to K and from mmHg to Pa)
        # Args:
        #   T: Temperature
        # Returns:
        #   p_sat: the saturation pressure of water 

        p_sat = (10**(8.07131 - 1730.63 / (T + 233.426 - 273.15))) * 133.3223684 

        if np.max(T) > 99+273 or np.min(T) < 273:
            raise Exception(f"Temperature out of valid scope for Antoine equation (0-99°C)! (Tmax = {np.max(T)-273.15}°C, Tmin = {np.min(T)-273.15}°C)")

        return p_sat